@@ -33,37 +33,11 @@
 #undef G_LOG_DOMAIN
 #define G_LOG_DOMAIN "libgsf:zip"
 
-typedef enum {
-	ZIP_STORED =          0,
-	ZIP_SHRUNK =          1,
-	ZIP_REDUCEDx1 =       2,
-	ZIP_REDUCEDx2 =       3,
-	ZIP_REDUCEDx3 =       4,
-	ZIP_REDUCEDx4 =       5,
-	ZIP_IMPLODED  =       6,
-	ZIP_TOKENIZED =       7,
-	ZIP_DEFLATED =        8,
-	ZIP_DEFLATED_BETTER = 9,
-	ZIP_IMPLODED_BETTER = 10
-} ZipCompressionMethod;
-
-typedef struct {	
-	char                 *name;
-	ZipCompressionMethod  compr_method;
-	guint32               crc32;
-	size_t                csize;
-	size_t                usize;
-	off_t                 offset;
-	off_t                 data_offset;
-
-	guint32               restlen;
-	guint32               crestlen;
-	z_stream              stream;
-} ZipDirent;
-
 typedef struct {
 	guint16   entries;
 	guint32   dir_pos;
+	GList     *dirent_list;
+	ZipVDir  *vdir;
 
 	int ref_count;
 } ZipInfo;
@@ -73,12 +47,16 @@ struct _GsfInfileZip {
 
 	GsfInput  *input;
 	ZipInfo	  *info;
-	GList     *dirent_list;
 
-	ZipDirent *dirent;
+	ZipVDir   *vdir;
 
+	z_stream  *stream;
+	guint32   restlen;
+	guint32   crestlen;
+	
 	guint8   *buf;
 	size_t    buf_size;
+	gsf_off_t seek_skipped;
 };
 
 typedef struct {
@@ -88,6 +66,57 @@ typedef struct {
 #define GSF_INFILE_ZIP_CLASS(k)    (G_TYPE_CHECK_CLASS_CAST ((k), GSF_INFILE_ZIP_TYPE, GsfInfileZipClass))
 #define GSF_IS_INFILE_ZIP_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GSF_INFILE_ZIP_TYPE))
 
+static ZipVDir *
+vdir_child_by_name (ZipVDir *vdir, char const * name)
+{
+	GSList *l;
+	ZipVDir *child;
+	
+	for (l = vdir->children; l; l = l->next) {
+		child = (ZipVDir *) l->data;
+		if (strcmp (child->name, name) == 0)
+			return child;
+	}
+	return NULL;
+}
+
+static ZipVDir *
+vdir_child_by_index (ZipVDir *vdir, int target)
+{
+	GSList *l;
+	
+	for (l = vdir->children; l; l = l->next)
+		if (target-- <= 0)
+			return (ZipVDir *) l->data;
+	return NULL;
+}
+
+static void
+vdir_insert (ZipVDir *vdir, char const * name, ZipDirent *dirent)
+{
+	const char *p;
+	char *dirname;
+	ZipVDir *child;
+	
+	p = strchr (name, ZIP_NAME_SEPARATOR);
+	if (p) {	/* A directory */
+		dirname = g_strndup (name, (gsize) (p - name));
+		child = vdir_child_by_name (vdir, dirname);
+		if (!child) {
+			child = vdir_new (dirname, TRUE, NULL);
+			vdir_add_child (vdir, child);
+		}
+		g_free (dirname);
+		if (*(p+1) != '\0') {
+			name = p+1;
+			vdir_insert (child, name, dirent);
+		}
+	} else { /* A simple file name */
+		child = vdir_new (name, FALSE, dirent);
+		vdir_add_child (vdir, child);
+	}
+}
+
 static gsf_off_t
 zip_find_trailer (GsfInfileZip *zip)
 {
@@ -154,7 +183,7 @@ zip_find_trailer (GsfInfileZip *zip)
 }
 
 static ZipDirent *
-zip_dirent_new (GsfInfileZip *zip, gsf_off_t *offset)
+zip_dirent_new_in (GsfInfileZip *zip, gsf_off_t *offset)
 {
 	static guint8 const dirent_signature[] =
 		{ 'P', 'K', 0x01, 0x02 };
@@ -181,16 +210,14 @@ zip_dirent_new (GsfInfileZip *zip, gsf_off_t *offset)
 	usize =         GSF_LE_GET_GUINT32 (data + ZIP_DIRENT_USIZE);
 	off =           GSF_LE_GET_GUINT32 (data + ZIP_DIRENT_OFFSET);
 
-	name = (gchar *) g_malloc ((gulong) (name_len + 1));
-	if ((data = gsf_input_read (zip->input, name_len, NULL)) == NULL) {
-		g_free (name);
+	if ((data = gsf_input_read (zip->input, name_len, NULL)) == NULL)
 		return NULL;
-	}
 
+	name = (gchar *) g_malloc ((gulong) (name_len + 1));
 	memcpy (name, data, name_len);
 	name[name_len] = '\0';
 
-	dirent = g_new0 (ZipDirent, 1);
+	dirent = zip_dirent_new ();
 	dirent->name = name;
 
 	dirent->compr_method =  compr_method;
@@ -198,27 +225,12 @@ zip_dirent_new (GsfInfileZip *zip, gsf_off_t *offset)
 	dirent->csize =         csize;
 	dirent->usize =         usize;
 	dirent->offset =        off;
-	dirent->restlen =       usize;
-	dirent->crestlen =      csize;
 
 	*offset += ZIP_DIRENT_SIZE + name_len + extras_len + comment_len;
 
 	return dirent;
 }
 
-static void
-zip_dirent_free (ZipDirent *dirent)
-{
-	g_return_if_fail (dirent != NULL);
-
-	(void) inflateEnd (&dirent->stream);
-	if (dirent->name != NULL) {
-		g_free (dirent->name);
-		dirent->name = NULL;
-	}
-	g_free (dirent);
-}
-
 /*****************************************************************************/
 static ZipInfo *
 zip_info_ref (ZipInfo *info)
@@ -230,9 +242,17 @@ zip_info_ref (ZipInfo *info)
 static void
 zip_info_unref (ZipInfo *info)
 {
+	GList *p;
+
 	if (info->ref_count-- != 1)
 		return;
 
+	vdir_free (info->vdir, FALSE);
+	for (p = info->dirent_list; p != NULL; p = p->next)
+		zip_dirent_free ((ZipDirent *) p->data);
+
+	g_list_free (info->dirent_list);
+
 	g_free (info);
 }
 
@@ -240,8 +260,6 @@ zip_info_unref (ZipInfo *info)
  * zip_dup :
  * @src :
  *
- * Utility routine to _partially_ replicate a file.  It does NOT init the dirent_list.
- *
  * Return value: the partial duplicate.
  **/
 static GsfInfileZip *
@@ -259,7 +277,7 @@ zip_dup (GsfInfileZip const *src)
 }
 
 /**
- * zip_init_info :
+ * zip_read_dirents:
  * @zip :
  * @err : optionally NULL
  *
@@ -269,7 +287,7 @@ zip_dup (GsfInfileZip const *src)
  * Return value: TRUE on error setting @err if it is supplied.
  **/
 static gboolean
-zip_init_info (GsfInfileZip *zip, GError **err)
+zip_read_dirents (GsfInfileZip *zip, GError **err)
 {
 	static guint8 const header_signature[] =
 		{ 'P', 'K', 0x03, 0x04 };
@@ -279,16 +297,6 @@ zip_init_info (GsfInfileZip *zip, GError **err)
 	ZipInfo *info;
 	gsf_off_t offset;
 
-	/* Check the header */
-	if (gsf_input_seek (zip->input, (gsf_off_t) 0, G_SEEK_SET) ||
-	    NULL == (header = gsf_input_read (zip->input, ZIP_HEADER_SIZE, NULL)) ||
-	    0 != memcmp (header, header_signature, sizeof (header_signature))) {
-		if (err != NULL)
-			*err = g_error_new (gsf_input_error (), 0,
-				"No Zip signature");
-		return TRUE;
-	}
-
 	/* Find and check the trailing header */
 	offset = zip_find_trailer (zip);
 	if (offset < 0) {
@@ -320,7 +328,7 @@ zip_init_info (GsfInfileZip *zip, GError **err)
 	for (i = 0, offset = dir_pos; i < entries; i++) {
 		ZipDirent *d;
 
-		d = zip_dirent_new (zip, &offset);
+		d = zip_dirent_new_in (zip, &offset);
 		if (d == NULL) {
 			if (err != NULL)
 				*err = g_error_new
@@ -329,7 +337,85 @@ zip_init_info (GsfInfileZip *zip, GError **err)
 			return TRUE;
 		}
 
-		zip->dirent_list = g_list_append (zip->dirent_list, d);
+		info->dirent_list = g_list_append (info->dirent_list, d);
+	}
+
+	return FALSE;
+}
+
+static void
+zip_build_vdirs (GsfInfileZip *zip)
+{
+	GList *l;
+	ZipDirent *dirent;
+	ZipInfo *info = zip->info;
+
+	info->vdir = vdir_new ("", TRUE, NULL);
+	for (l = info->dirent_list; l; l = l->next) {
+		dirent = (ZipDirent *) l->data;
+		vdir_insert (info->vdir, dirent->name, dirent);
+	}
+}
+
+/**
+ * zip_init_info :
+ * @zip :
+ * @err : optionally NULL
+ *
+ * Read zip headers and do some sanity checking
+ * along the way.
+ *
+ * Return value: TRUE on error setting @err if it is supplied.
+ **/
+static gboolean
+zip_init_info (GsfInfileZip *zip, GError **err)
+{
+	gboolean ret;
+	
+	ret = zip_read_dirents (zip, err);
+	if (ret != FALSE)
+		return ret;
+	zip_build_vdirs (zip);
+
+	return FALSE;
+}
+
+static gboolean
+zip_child_init (GsfInfileZip *child)
+{
+	static guint8 const header_signature[] =
+		{ 'P', 'K', 0x03, 0x04 };
+	guint8 const *data;
+	guint16 name_len, extras_len;
+	ZipDirent *dirent = child->vdir->dirent;
+
+	/* skip local header
+	 * should test tons of other info, but trust that those are correct
+	 **/
+
+	if (gsf_input_seek (child->input, (gsf_off_t) dirent->offset,
+			    G_SEEK_SET) ||
+	    NULL == (data = gsf_input_read (child->input, ZIP_FILE_HEADER_SIZE, NULL)) ||
+	    0 != memcmp (data, header_signature, sizeof (header_signature))) {
+		return TRUE;
+	}
+
+	name_len =   GSF_LE_GET_GUINT16 (data + ZIP_FILE_HEADER_NAME_SIZE);
+	extras_len = GSF_LE_GET_GUINT16 (data + ZIP_FILE_HEADER_EXTRAS_SIZE);
+
+	dirent->data_offset = dirent->offset + ZIP_FILE_HEADER_SIZE + name_len + extras_len;
+	child->restlen  = dirent->usize;
+	child->crestlen = dirent->csize;
+
+	if (dirent->compr_method != ZIP_STORED) {
+		int err;
+
+		if (!child->stream) {
+			child->stream = g_new0 (z_stream, 1);
+		}
+		err = inflateInit2 (child->stream, -MAX_WBITS);
+		if (err != Z_OK)
+			return TRUE;
 	}
 
 	return FALSE;
@@ -349,32 +435,40 @@ gsf_infile_zip_dup (GsfInput *src_input, GError **err)
 					    "Something went wrong in zip_dup.");		return NULL;
 	}
 
-	dst->dirent = src->dirent;
+	dst->vdir = src->vdir;
+
+	if (dst->vdir->dirent)
+		if (zip_child_init (dst) != FALSE) {
+			g_object_unref (dst);
+			if (err != NULL)
+				*err = g_error_new (gsf_input_error (), 0,
+						    "Something went wrong in zip_child_init.");
+			return NULL;
+		}
 
 	return GSF_INPUT (dst);
 }
 
 static gboolean
-zip_update_stream_in (GsfInfileZip const *zip)
+zip_update_stream_in (GsfInfileZip *zip)
 {
-	ZipDirent *dirent;
 	guint32 read_now;
 	guint8 const *data;
-
-	dirent = zip->dirent;
-
-	if (dirent->crestlen == 0)
+	gsf_off_t pos;
+	
+	if (zip->crestlen == 0)
 		return FALSE;
 
-	read_now = (dirent->crestlen > ZIP_BLOCK_SIZE) ? ZIP_BLOCK_SIZE : dirent->crestlen;
+	read_now = MIN (zip->crestlen, ZIP_BLOCK_SIZE);
 
-	gsf_input_seek (zip->input, (gsf_off_t) (dirent->data_offset + dirent->stream.total_in), G_SEEK_SET);
+	pos = zip->vdir->dirent->data_offset + zip->stream->total_in;
+	gsf_input_seek (zip->input, pos, G_SEEK_SET);
 	if ((data = gsf_input_read (zip->input, read_now, NULL)) == NULL)
 		return FALSE;
 
-	dirent->crestlen -= read_now;
-	dirent->stream.next_in  = (unsigned char *) data;      /* next input byte */
-	dirent->stream.avail_in = read_now;  /* number of bytes available at next_in */
+	zip->crestlen -= read_now;
+	zip->stream->next_in  = (unsigned char *) data;      /* next input byte */
+	zip->stream->avail_in = read_now;  /* number of bytes available at next_in */
 
 	return TRUE;
 }
@@ -383,15 +477,18 @@ static guint8 const *
 gsf_infile_zip_read (GsfInput *input, size_t num_bytes, guint8 *buffer)
 {
 	GsfInfileZip *zip = GSF_INFILE_ZIP (input);
-	ZipDirent    *dirent = zip->dirent;
+	ZipVDir      *vdir = zip->vdir;
+	gsf_off_t pos;
 
-	if (dirent->restlen < num_bytes)
+	if (zip->restlen < num_bytes)
 		return NULL;
 
-	switch (dirent->compr_method) {
+	switch (vdir->dirent->compr_method) {
 	case ZIP_STORED:
-		dirent->restlen -= num_bytes;
-		return gsf_input_read (input, num_bytes, buffer);
+		zip->restlen -= num_bytes;
+		pos = zip->vdir->dirent->data_offset + input->cur_offset;
+		gsf_input_seek (zip->input, pos, G_SEEK_SET);
+		return gsf_input_read (zip->input, num_bytes, buffer);
 		break;
 	case ZIP_DEFLATED:
 
@@ -405,28 +502,28 @@ gsf_infile_zip_read (GsfInput *input, size_t num_bytes, guint8 *buffer)
 			buffer = zip->buf;
 		}
 
-		dirent->stream.avail_out = num_bytes;
-		dirent->stream.next_out = (unsigned char *)buffer;
+		zip->stream->avail_out = num_bytes;
+		zip->stream->next_out = (unsigned char *)buffer;
 
 		do {
 			int err;
 			int startlen;
 
-			if (dirent->crestlen > 0 && dirent->stream.avail_in == 0)
+			if (zip->crestlen > 0 && zip->stream->avail_in == 0)
 				if (!zip_update_stream_in (zip))
 					break;
 
-			startlen = dirent->stream.total_out;
-			err = inflate(&dirent->stream, Z_NO_FLUSH);
+			startlen = zip->stream->total_out;
+			err = inflate(zip->stream, Z_NO_FLUSH);
 
 			if (err == Z_STREAM_END) 
-				dirent->restlen = 0;
+				zip->restlen = 0;
 			else if (err == Z_OK)
-				dirent->restlen -= (dirent->stream.total_out - startlen);
+				zip->restlen -= (zip->stream->total_out - startlen);
 			else
 				break;
 
-		} while (dirent->restlen && dirent->stream.avail_out);
+		} while (zip->restlen && zip->stream->avail_out);
 
 		return buffer;
 		break;
@@ -442,10 +539,38 @@ static gboolean
 gsf_infile_zip_seek (GsfInput *input, gsf_off_t offset, GSeekType whence)
 {
 	GsfInfileZip *zip = GSF_INFILE_ZIP (input);
+	/* Global flag -- we don't want one per stream.  */
+	static gboolean warned = FALSE;
+	gsf_off_t pos = offset;
 
-	(void) zip;
-	(void) offset; 
-	(void) whence;
+	/* Note, that pos has already been sanity checked.  */
+	switch (whence) {
+	case G_SEEK_SET : break;
+	case G_SEEK_CUR : pos += input->cur_offset;	break;
+	case G_SEEK_END : pos += input->size;		break;
+	default : return TRUE;
+	}
+
+	if (zip->stream) {
+		zip->stream->next_in  = NULL;
+		zip->stream->avail_in = 0;
+		zip->stream->total_in = 0;
+	}
+
+	if (zip_child_init (zip) != FALSE)
+		return TRUE;
+
+	input->cur_offset = 0;
+	if (gsf_input_seek_emulate (input, pos))
+		return TRUE;
+
+	zip->seek_skipped += pos;
+	if (!warned &&
+	    zip->seek_skipped != pos && /* Don't warn for single seek.  */
+	    zip->seek_skipped >= 1000000) {
+		warned = TRUE;
+		g_warning ("Seeking in zip child streams is awfully slow.");
+	}
 
 	return FALSE;
 }
@@ -456,54 +581,27 @@ gsf_infile_zip_seek (GsfInput *input, gsf_off_t offset, GSeekType whence)
 
 
 static GsfInput *
-gsf_infile_zip_new_child (GsfInfileZip *parent, ZipDirent *dirent)
+gsf_infile_zip_new_child (GsfInfileZip *parent, ZipVDir *vdir)
 {
 	GsfInfileZip *child;
-	static guint8 const header_signature[] =
-		{ 'P', 'K', 0x03, 0x04 };
-	guint8 const *data;
-	guint16 name_len, extras_len;
-
+	ZipDirent *dirent = vdir->dirent;
 	child = zip_dup (parent);
 
-	gsf_input_set_size (GSF_INPUT (child), (gsf_off_t) dirent->usize);
-	gsf_input_set_name (GSF_INPUT (child), dirent->name);
+	gsf_input_set_name (GSF_INPUT (child), vdir->name);
 	gsf_input_set_container (GSF_INPUT (child), GSF_INFILE (parent));
 
-	child->dirent = dirent;
+	child->vdir = vdir;
 
-	/* skip local header
-	 * should test tons of other info, but trust that those are correct
-	 **/
-
-	if (gsf_input_seek (child->input, (gsf_off_t) dirent->offset,
-			    G_SEEK_SET) ||
-	    NULL == (data = gsf_input_read (child->input, ZIP_FILE_HEADER_SIZE, NULL)) ||
-	    0 != memcmp (data, header_signature, sizeof (header_signature))) {
-		g_object_unref (child);
-		return NULL;
-	}
-
-	name_len =   GSF_LE_GET_GUINT16 (data + ZIP_FILE_HEADER_NAME_SIZE);
-	extras_len = GSF_LE_GET_GUINT16 (data + ZIP_FILE_HEADER_EXTRAS_SIZE);
-
-	dirent->data_offset = dirent->offset + ZIP_FILE_HEADER_SIZE + name_len + extras_len;
-
-	if (dirent->compr_method != ZIP_STORED) {
-		int err;
-
-		if (!zip_update_stream_in (child)) {
-			g_object_unref (child);
-			return NULL;
-		}
-
-		err = inflateInit2 (&child->dirent->stream, -MAX_WBITS);
-		if (err != Z_OK) {
+	if (dirent) {
+		gsf_input_set_size (GSF_INPUT (child),
+				    (gsf_off_t) dirent->usize);
+		if (zip_child_init (child) != FALSE) {
 			g_object_unref (child);
 			return NULL;
 		}
+	} else {
+		gsf_input_set_size (GSF_INPUT (child), (gsf_off_t) 0);
 	}
-
 	return GSF_INPUT (child);
 }
 
@@ -511,11 +609,11 @@ static GsfInput *
 gsf_infile_zip_child_by_index (GsfInfile *infile, int target)
 {
 	GsfInfileZip *zip = GSF_INFILE_ZIP (infile);
-	GList *p;
+	ZipVDir *child_vdir = vdir_child_by_index (zip->vdir, target);
+
+	if (child_vdir)
+		return gsf_infile_zip_new_child (zip, child_vdir);
 
-	for (p = zip->dirent_list; p != NULL; p = p->next)
-		if (target-- <= 0)
-			return gsf_infile_zip_new_child (zip, p->data);
 	return NULL;
 }
 
@@ -523,11 +621,11 @@ static char const *
 gsf_infile_zip_name_by_index (GsfInfile *infile, int target)
 {
 	GsfInfileZip *zip = GSF_INFILE_ZIP (infile);
-	GList *p;
+	ZipVDir *child_vdir = vdir_child_by_index (zip->vdir, target);
+
+	if (child_vdir)
+		return child_vdir->name;
 
-	for (p = zip->dirent_list; p != NULL; p = p->next)
-		if (target-- <= 0)
-			return ((ZipDirent *)p->data)->name;
 	return NULL;
 }
 
@@ -535,13 +633,11 @@ static GsfInput *
 gsf_infile_zip_child_by_name (GsfInfile *infile, char const *name)
 {
 	GsfInfileZip *zip = GSF_INFILE_ZIP (infile);
-	GList *p;
+	ZipVDir *child_vdir = vdir_child_by_name (zip->vdir, name);
+
+	if (child_vdir)
+		return gsf_infile_zip_new_child (zip, child_vdir);
 
-	for (p = zip->dirent_list; p != NULL; p = p->next) {
-		ZipDirent *dirent = p->data;
-		if (dirent->name != NULL && !strcmp (name, dirent->name))
-			return gsf_infile_zip_new_child (zip, p->data);
-	}
 	return NULL;
 }
 
@@ -550,7 +646,11 @@ gsf_infile_zip_num_children (GsfInfile *infile)
 {
 	GsfInfileZip *zip = GSF_INFILE_ZIP (infile);
 
-	return g_list_length (zip->dirent_list);
+	g_return_val_if_fail (zip->vdir != NULL, -1);
+
+	if (!zip->vdir->is_directory)
+		return -1;
+	return g_slist_length (zip->vdir->children);
 }
 
 static void
@@ -558,7 +658,6 @@ gsf_infile_zip_finalize (GObject *obj)
 {
 	GObjectClass *parent_class;
 	GsfInfileZip *zip = GSF_INFILE_ZIP (obj);
-	GList *p;
 
 	if (zip->input != NULL) {
 		g_object_unref (G_OBJECT (zip->input));
@@ -568,11 +667,10 @@ gsf_infile_zip_finalize (GObject *obj)
 		zip_info_unref (zip->info);
 		zip->info = NULL;
 	}
-	for (p = zip->dirent_list; p != NULL; p = p->next)
-		zip_dirent_free ((ZipDirent *) p->data);
-
-	g_list_free (zip->dirent_list);
-
+	if (zip->stream)
+		(void) inflateEnd (zip->stream);
+	g_free (zip->stream);
+	
 	g_free (zip->buf);
 
 	parent_class = g_type_class_peek (GSF_INFILE_TYPE);
@@ -586,7 +684,13 @@ gsf_infile_zip_init (GObject *obj)
 	GsfInfileZip *zip = (GsfInfileZip *)obj;
 	zip->input =   NULL;
 	zip->info  =   NULL;
-	zip->dirent =  NULL;
+	zip->vdir  =   NULL;
+	zip->stream   =  NULL;
+	zip->restlen  = 0;
+	zip->crestlen = 0;
+	zip->buf   =   NULL;
+	zip->buf_size = 0;
+	zip->seek_skipped = 0;
 }
 
 static void
@@ -635,8 +739,7 @@ gsf_infile_zip_new (GsfInput *source, GError **err)
 		g_object_unref (G_OBJECT (zip));
 		return NULL;
 	}
-	zip->buf		= NULL;
-	zip->buf_size		= 0;
+	zip->vdir               = zip->info->vdir;
 
 	return GSF_INFILE (zip);
 }
